/*  Copyright [2020] [Verdant Robotics]

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
*/

#include "nvmm_encoder.h"

#include <linux/videodev2.h>
#include <nvbuf_utils.h>
#include <nvToolsExt.h>

#define MINIMP4_IMPLEMENTATION
#include "minimp4.h"

constexpr uint32_t MAX_OUT_BYTES = 2 * 1024 * 1024;
constexpr uint32_t INPUT_BUFFERS = 1;
constexpr uint32_t OUTPUT_BUFFERS = 1;

#define CUDA_CALL(call)                                                                                     \
  {                                                                                                         \
    CUresult res = call;                                                                                    \
    if (CUDA_SUCCESS != res) {                                                                              \
      const char* errName = nullptr;                                                                        \
      const char* errStr = nullptr;                                                                         \
      cuGetErrorName(res, &errName);                                                                        \
      cuGetErrorString(res, &errStr);                                                                       \
      printf("NVMME CUDA API failure %s: %s. Source: %s @ %s:%d\n", errName, errStr, #call,                 \
                 __FILE__, __LINE__);                                                                        \
      return false;                                                                                         \
    }                                                                                                       \
  }                                                                                                         \
  ((void)0)

#define CALL_OR_FALSE(call)                                                                          \
  {                                                                                                  \
    const int res = call;                                                                            \
    if (0 != res) {                                                                                  \
      printf("NVMME error %d. Source: %s @ %s:%d\n", res, #call, __FILE__, __LINE__);                \
      return false;                                                                                  \
    }                                                                                                \
  }                                                                                                  \
  ((void)0)

static int WriteCallback(int64_t offset, const void* buffer, size_t size, void* token) {
  std::FILE* f = reinterpret_cast<std::FILE*>(token);
  std::fseek(f, offset, SEEK_SET);
  return std::fwrite(buffer, 1, size, f) != size;
}

static bool EncoderCapturePlaneDequeued(v4l2_buffer* v4l2Buffer, NvBuffer* buffer, NvBuffer* sharedBuffer,
                                        void* arg) {
  (void)sharedBuffer;
  auto* nvmmEncoder = reinterpret_cast<libvideo::NvMmEncoder*>(arg);
  auto* encoder = nvmmEncoder->encoder();
  const std::string_view name = nvmmEncoder->name();

  if (!v4l2Buffer) {
    printf("NVMME %s failed to dequeue buffer from encoder capture plane\n", name.data());
    nvmmEncoder->deInit();
    return false;
  }

  const uint32_t frameNum = encoder->capture_plane.getTotalDequeuedBuffers() - 1;

  // Write the encoded data to our output file
  const u_char* data = buffer->planes[0].data;
  const int nalSize = int(buffer->planes[0].bytesused);
  const uint32_t timestamp90kHz = 90000 / nvmmEncoder->fps();
  auto* writer = reinterpret_cast<mp4_h26x_writer_t*>(nvmmEncoder->mp4Writer());
  // printf("NVMME %s writing %d bytes for frame %u to %s\n", name.data(), nalSize, frameNum,
  //             nvmmEncoder->filename().data());
  if (MP4E_STATUS_OK != mp4_h26x_write_nal(writer, data, nalSize, timestamp90kHz)) {
    printf("mp4_h26x_write_nal(%p, %p, %d, 90000/%zu)\n", nvmmEncoder->mp4Writer(),
               reinterpret_cast<const void*>(data), nalSize, nvmmEncoder->fps());
  }

  // Re-enqueue this buffer on the capture plane now that we're done with it
  // vlog_finest(VCAT_VID, "NVMME capture_plane.qBuffer()");
  const int res = encoder->capture_plane.qBuffer(*v4l2Buffer, nullptr);
  if (0 != res) {
    printf("NVMME %s error %d requeuing buffer to the capture plane\n", name.data(), res);
    nvmmEncoder->deInit();
    return false;
  }

  // Check for an end of stream signal
  if (buffer->planes[0].bytesused == 0) {
    printf("NVMME %s write callback received EOS signal\n", name.data());
    return false;
  }

  return true;
}

static uint32_t GetH264Bitrate(uint32_t width, uint32_t height, uint32_t fps) {
  constexpr uint32_t MAX_SLOW_H264_BITRATE = 70000000;                                             // [bit/s]
  constexpr uint32_t MAX_FAST_H264_BITRATE = 64000000;                                             // [bit/s]
  constexpr uint32_t MAX_ULTRAFAST_H264_BITRATE = 48000000;                                        // [bit/s]
  constexpr double SLOW_BPP = double(MAX_SLOW_H264_BITRATE) / (4096.0 * 2160.0 * 30.0);            // [bpp]
  constexpr double FAST_BPP = double(MAX_FAST_H264_BITRATE) / (4096.0 * 2160.0 * 30.0);            // [bpp]
  constexpr double ULTRAFAST_BPP = double(MAX_ULTRAFAST_H264_BITRATE) / (4096.0 * 2160.0 * 30.0);  // [bpp]

  return std::min(uint32_t(double(width * height * fps) * FAST_BPP), MAX_FAST_H264_BITRATE);
}

namespace libvideo {

struct NvtxRange {
  NvtxRange() = delete;
  explicit NvtxRange(const std::string_view name) { nvtxRangePush(name.data()); }
  ~NvtxRange() { nvtxRangePop(); }
};

bool NvMmEncoder::init(const std::string_view name, size_t width, size_t height, size_t fps,
                       CUdevice cudaDevice) {
  printf("NVMME init(%s, %zu, %zu, %zu, %d)\n", name.data(), width, height, fps,
             cudaDevice);

  if (encoder_) deInit();

  name_ = name;
  width_ = width;
  height_ = height;
  fps_ = fps;

  CUDA_CALL(cuInit(0));

  // Find the first CUDA device if we were not given a device reference
  if (cudaDevice == CU_DEVICE_INVALID) {
    CUDA_CALL(cuDeviceGet(&cudaDevice, 0));
  }

  // Retain a handle to the primary context for the specified CUDA device
  cudaDevice_ = cudaDevice;
  CUDA_CALL(cuDevicePrimaryCtxRetain(&cudaContext_, cudaDevice_));

  // Instantiate a video encoder
  encoder_ = NvVideoEncoder::createVideoEncoder(name_.c_str());
  if (!encoder_) {
    printf("NVMME createVideoEncoder(\"%s\") failed\n", name_.c_str());
    assert(false);
    return false;
  }

  printf("NVMME init succeeded for %s\n", name.data());
  return true;
}

void NvMmEncoder::deInit() {
  printf("NVMME %s deInit()\n", name_.c_str());

  endEncoding();

  if (encoder_) {
    encoder_->output_plane.deinitPlane();
    encoder_->capture_plane.deinitPlane();
    delete encoder_;
    encoder_ = nullptr;
  }

  if (cudaContext_) {
    CUresult result = cuDevicePrimaryCtxRelease(cudaDevice_);
    if (result != CUDA_SUCCESS) {
      printf("NVMME %s cuDevicePrimaryCtxRelease(%d) failed: %d\n", name_.c_str(), cudaDevice_,
                   result);
    }
    cudaContext_ = nullptr;
    cudaDevice_ = CU_DEVICE_INVALID;
  }

  filename_.clear();
  fps_ = 0;
}

bool NvMmEncoder::beginEncoding(std::string_view filename, std::string_view message) {
  if (running_) endEncoding();

  frameIndex_ = 0;

  // Open the output file for binary writing
  filename_ = filename;
  outputFilePtr_ = std::fopen(filename.data(), "wb");
  if (!outputFilePtr_ || std::ferror(outputFilePtr_)) {
    printf("NVMME %s failed to open \"%s\" for writing\n", name_.c_str(), filename.data());
    return false;
  }

  printf("Starting NVMME %s video encoding to %s\n", name_.c_str(), filename.data());

  // Initialize the MP4 writer
  constexpr int IS_HEVC = 0;
  auto muxer = MP4E_open(0, 0, outputFilePtr_, WriteCallback);
  auto mp4Writer = new mp4_h26x_writer_t;
  muxer_ = muxer;
  mp4Writer_ = mp4Writer;
  if (MP4E_STATUS_OK != mp4_h26x_write_init(mp4Writer, muxer, int(width_), int(height_), IS_HEVC)) {
    printf("mp4_h26x_write_init(%p, %p, %zu, %zu, %d)\n", mp4Writer_, muxer_, width_, height_, IS_HEVC);
    return false;
  }
  if (!message.empty()) {
    MP4E_set_text_comment(muxer, message.data());
  }

  const auto w = uint32_t(width_);
  const auto h = uint32_t(height_);
  const auto fps = uint32_t(fps_);

  // Configure the video encoder
  CALL_OR_FALSE(encoder_->setCapturePlaneFormat(V4L2_PIX_FMT_H264, w, h, MAX_OUT_BYTES));
  CALL_OR_FALSE(encoder_->setOutputPlaneFormat(V4L2_PIX_FMT_YUV420M, w, h));
  CALL_OR_FALSE(encoder_->setMaxPerfMode(1));
  CALL_OR_FALSE(encoder_->setHWPresetType(V4L2_ENC_HW_PRESET_FAST));
  CALL_OR_FALSE(encoder_->setProfile(V4L2_MPEG_VIDEO_H264_PROFILE_MAIN));
  CALL_OR_FALSE(encoder_->setIFrameInterval(uint32_t(fps_)));
  CALL_OR_FALSE(encoder_->setFrameRate(uint32_t(fps_), 1));
  CALL_OR_FALSE(encoder_->setBitrate(GetH264Bitrate(w, h, fps)));
  CALL_OR_FALSE(encoder_->output_plane.setupPlane(V4L2_MEMORY_MMAP, INPUT_BUFFERS, true, false));
  CALL_OR_FALSE(encoder_->capture_plane.setupPlane(V4L2_MEMORY_MMAP, OUTPUT_BUFFERS, true, false));
  encoder_->capture_plane.setDQThreadCallback(EncoderCapturePlaneDequeued);
  CALL_OR_FALSE(encoder_->output_plane.setStreamStatus(true));
  CALL_OR_FALSE(encoder_->capture_plane.setStreamStatus(true));
  encoder_->capture_plane.startDQThread(this);

  // Enqueue all the empty capture plane buffers
  for (uint32_t i = 0; i < encoder_->capture_plane.getNumBuffers(); i++) {
    v4l2_buffer v4l2Buffer{};
    std::array<v4l2_plane, MAX_PLANES> planes{};
    v4l2Buffer.index = i;
    v4l2Buffer.m.planes = planes.data();
    CALL_OR_FALSE(encoder_->capture_plane.qBuffer(v4l2Buffer, nullptr));
  }

  printf("NVMME %s beginEncoding succeeded for %s\n", name_.c_str(), filename.data());
  running_ = true;
  return true;
}


bool NvMmEncoder::pushFrame(const Image& image) {
  constexpr uint32_t NUM_RETRIES = 1;

  if (!running_) {
    printf("NVMME %s pushFrame called when it is not running\n", name_.c_str());
    return false;
  }

  NvtxRange{"pushFrame"};

  const bool initBuffers = frameIndex_ < encoder_->output_plane.getNumBuffers();
  nvtxRangePush(initBuffers ? "output_plane.getNthBuffer" : "output_plane.dqBuffer");
  v4l2_buffer v4l2Buffer{};
  std::array<v4l2_plane, MAX_PLANES> planes{};
  NvBuffer* buffer = nullptr;

  v4l2Buffer.m.planes = planes.data();

  if (initBuffers) {
    // printf("NVMME %s output_plane.getNthBuffer(%zu)\n", name_.c_str(), frameIndex_);
    buffer = encoder_->output_plane.getNthBuffer(uint32_t(frameIndex_));
    v4l2Buffer.index = uint32_t(frameIndex_);
  } else {
    CALL_OR_FALSE(encoder_->output_plane.dqBuffer(v4l2Buffer, &buffer, nullptr, NUM_RETRIES));
    // printf("NVMME %s output_plane.dqBuffer(), index=%u, frameIndex=%zu\n", name_.c_str(),
    //             v4l2Buffer.index, frameIndex_);
  }
  nvtxRangePop();

  const size_t luminaStride = buffer->planes[0].fmt.stride;
  const size_t luminaHeight = buffer->planes[0].fmt.height;
  const size_t chromaStride = buffer->planes[1].fmt.stride;
  const size_t chromaHeight = buffer->planes[1].fmt.height;

  const size_t srcPitch = size_t(image.pitch);
  const size_t srcWidth = size_t(image.width);
  const size_t srcHeight = size_t(image.height);

  NvtxRange{"cuMemcpy2D"};
  CUDA_CALL(cuCtxPushCurrent(cudaContext_));

  // Y plane
  CUDA_MEMCPY2D m{};
  m.srcMemoryType = CU_MEMORYTYPE_DEVICE;
  m.srcDevice = CUdeviceptr(image.cuda_handle);
  m.srcPitch = srcPitch;
  m.dstMemoryType = CU_MEMORYTYPE_HOST;
  m.dstHost = reinterpret_cast<void*>(buffer->planes[0].data);
  m.dstPitch = luminaStride;
  m.WidthInBytes = srcWidth;
  m.Height = srcHeight;
  CUDA_CALL(cuMemcpy2D(&m));

  // U plane
  m.srcDevice = CUdeviceptr(image.cuda_handle + srcPitch * srcHeight);
  m.srcPitch = (srcPitch + 1) / 2;
  m.dstHost = reinterpret_cast<void*>(buffer->planes[1].data);
  m.dstPitch = chromaStride;
  m.WidthInBytes = (srcWidth + 1) / 2;
  m.Height = (srcHeight + 1) / 2;
  CUDA_CALL(cuMemcpy2D(&m));

  // V plane
  m.srcDevice = CUdeviceptr(image.cuda_handle + srcPitch * srcHeight + m.srcPitch * m.Height);
  m.dstHost = reinterpret_cast<void*>(buffer->planes[2].data);
  CUDA_CALL(cuMemcpy2D(&m));

  CUDA_CALL(cuCtxPopCurrent(nullptr));

  buffer->planes[0].bytesused = uint32_t(luminaStride * luminaHeight);
  buffer->planes[1].bytesused = uint32_t(chromaStride * chromaHeight);
  buffer->planes[2].bytesused = uint32_t(chromaStride * chromaHeight);

  // Enqueue this frame for video encoding
  {
    NvtxRange{"output_plane.qBuffer"};
    CALL_OR_FALSE(encoder_->output_plane.qBuffer(v4l2Buffer, nullptr));
  }

  ++frameIndex_;
  return true;
}

bool NvMmEncoder::endEncoding() {
  constexpr uint32_t NUM_RETRIES = 10;
  constexpr uint32_t MAX_WAIT_MS = 2000;

  if (!filename_.empty()) {
    printf("NVMME %s endEncoding() for %s\n", name_.c_str(), filename_.c_str());
  }

  if (running_ ) {
    // Send an "End Of Signal (EOS)" command to the video encoder if it's currently running
    v4l2_buffer v4l2Buffer{};
    std::array<v4l2_plane, MAX_PLANES> planes{};

    v4l2Buffer.m.planes = planes.data();
    v4l2Buffer.m.planes[0].bytesused = 0;

    if (frameIndex_ < encoder_->output_plane.getNumBuffers()) {
      v4l2Buffer.index = uint32_t(frameIndex_);
    } else {
      NvBuffer* buffer;
      int ret = encoder_->output_plane.dqBuffer(v4l2Buffer, &buffer, nullptr, NUM_RETRIES);
      if (0 != ret) {
        printf("NVMME %s error %d dequeueing EOS frame buffer %zu\n", name_.c_str(), ret,
                   frameIndex_);
        assert(false);
        return false;
      }
    }

    int ret = encoder_->output_plane.qBuffer(v4l2Buffer, nullptr);
    if (0 != ret) {
      printf("NVMME %s error %d queueing EOS frame buffer %zu\n", name_.c_str(), ret,
                 frameIndex_);
      assert(false);
      return false;
    }

    // Wait for encoding to finish then destroy output/capture streams
    CALL_OR_FALSE(encoder_->capture_plane.waitForDQThread(MAX_WAIT_MS));
    encoder_->output_plane.deinitPlane();
    encoder_->capture_plane.deinitPlane();
  }

  running_ = false;

  // Destroy the MP4 muxer
  if (muxer_) {
    MP4E_close(reinterpret_cast<MP4E_mux_t*>(muxer_));
    muxer_ = nullptr;
  }

  if (mp4Writer_) {
    auto mp4Writer = reinterpret_cast<mp4_h26x_writer_t*>(mp4Writer_);
    mp4_h26x_write_close(mp4Writer);
    delete mp4Writer;
    mp4Writer_ = nullptr;
  }

  // Close the output file
  if (outputFilePtr_) {
    std::fclose(outputFilePtr_);
    outputFilePtr_ = nullptr;
  }

  if (!filename_.empty()) {
    printf("Finished NVMME %s video encoding to %s\n", name_.c_str(), filename_.c_str());
  }
  return true;
}

}  // namespace libvideo
